#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <sys/types.h>		/* IEEE Std 1003.1-2001 non-conformance. */
#include <regex.h>
#include <setjmp.h>

#include "config.h"
#include "imapfilter.h"
#include "buffer.h"


extern conn_t connpri, connaux;
extern jmp_buf acctloop;

buffer_t ibuf;			/* Input buffer. */


void receive_response(conn_t * conn, char *buf);
void bye_response(char *buf);
int analyze_response(conn_t * conn, char *buf);


/*
 * Read one packet of data that the server sent.
 */
void
receive_response(conn_t * conn, char *buf)
{
	if (socket_read(conn, buf) == ERROR_NETWORK)
		longjmp(acctloop, -1);

#ifdef DEBUG
	fprintf(stderr, "debug: getting response (%s):\n\n%s\n",
	    (conn == &connpri ? "primary" : "auxiliary"), buf);
#endif
}


/*
 * Get server response to client's request.
 */
int
server_response(conn_t * conn, unsigned int tag)
{
	reset_buffer(&ibuf);

	do {
		check_buffer(&ibuf, strlen(ibuf.data) + RESPONSE_BUF);
		receive_response(conn, ibuf.data + strlen(ibuf.data));
		bye_response(ibuf.data);
	} while (!strcasestr(ibuf.data, ultostr(tag, 16)));

	return analyze_response(conn, ibuf.data);
}


/*
 * Check if server sent a BYE response (connection is closed immediately).
 */
void
bye_response(char *buf)
{
	if (strcasestr(buf, "* BYE"))
		longjmp(acctloop, -1);
}


/*
 * Process the greeting that server sends during connection.
 */
int
greeting_response(conn_t * conn)
{
	reset_buffer(&ibuf);

	receive_response(conn, ibuf.data);

	verbose("%s: %s", (conn == &connpri ? "S" : "s"), ibuf);

	bye_response(ibuf.data);

	if (strcasestr(ibuf.data, "* PREAUTH"))
		return RESPONSE_PREAUTH;

	return RESPONSE_OK;
}


/*
 * Process the data that server sent due to IMAP LOGOUT client request.
 */
int
logout_response(conn_t * conn, unsigned int tag)
{
	reset_buffer(&ibuf);

	do {
		check_buffer(&ibuf, strlen(ibuf.data) + RESPONSE_BUF);
		receive_response(conn, ibuf.data + strlen(ibuf.data));
	} while (!strcasestr(ibuf.data, ultostr(tag, 16)));

	return analyze_response(conn, ibuf.data);
}


/*
 * Process the data that server sent due to IMAP CAPABILITY client request.
 */
int
capability_response(conn_t * conn, unsigned int tag)
{
	reset_buffer(&ibuf);

	do {
		check_buffer(&ibuf, strlen(ibuf.data) + RESPONSE_BUF);
		receive_response(conn, ibuf.data + strlen(ibuf.data));
		bye_response(ibuf.data);
	} while (!strcasestr(ibuf.data, ultostr(tag, 16)));

	if (!strcasestr(ibuf.data, "IMAP4rev1")) {
		error("server does not support IMAP4rev1 protocol\n");
		return -2;
	}
	if (strcasestr(ibuf.data, "NAMESPACE"))
		conn->caps |= CAPABILITY_NAMESPACE;
#ifdef CRAM_MD5
	if (strcasestr(ibuf.data, "AUTH=CRAM-MD5"))
		conn->caps |= CAPABILITY_AUTH_CRAM_MD5;
#endif
#ifdef SSL_TLS
	if (strcasestr(ibuf.data, "STARTTLS"))
		conn->caps |= CAPABILITY_STARTTLS;
#endif
	return analyze_response(conn, ibuf.data);
}


#ifdef CRAM_MD5
/*
 * Process the data that server sent due to IMAP AUTHENTICATE client request.
 */
int
authenticate_response(conn_t * conn, unsigned int tag, unsigned char **cont)
{
	int i;
	char *c;

	reset_buffer(&ibuf);

	do {
		check_buffer(&ibuf, strlen(ibuf.data) + RESPONSE_BUF);
		receive_response(conn, ibuf.data + strlen(ibuf.data));
		bye_response(ibuf.data);
	} while (strlen(ibuf.data) == RESPONSE_BUF &&
	    !strcasestr(ibuf.data, ultostr(tag, 16)));

	if (cont != NULL && ibuf.data[0] == '+' && ibuf.data[1] == ' ') {
		c = *cont = (unsigned char *)xmalloc(strlen(ibuf.data) + 1);

		for (i = 2; ibuf.data[i] != '\r'; i++)
			*c++ = ibuf.data[i];

		*c = '\0';
	}
	return analyze_response(conn, ibuf.data);
}
#endif


/*
 * Process the data that server sent due to IMAP NAMESPACE client request.
 */
int
namespace_response(conn_t * conn, unsigned int tag)
{
	char *c, *d;

	reset_buffer(&ibuf);

	do {
		check_buffer(&ibuf, strlen(ibuf.data) + RESPONSE_BUF);
		receive_response(conn, ibuf.data + strlen(ibuf.data));
		bye_response(ibuf.data);
	} while (!strcasestr(ibuf.data, ultostr(tag, 16)));

	if ((c = strcasestr(ibuf.data, "* NAMESPACE"))) {
		c += 12;
		if (strncasecmp(c, "NIL", 3)) {
			c = strchr(c, '"') + 1;
			d = strchr(c, '"') + 1;

			strncat(conn->nsp.prefix, c, d - c - 1);
			conn->nsp.delim = *(strchr(d, '"') + 1);
		}
	}
#ifdef DEBUG
	fprintf(stderr, "debug: namespace (%s): '%s' '%c'\n",
	    (conn == &connpri ? "primary" : "auxiliary"), conn->nsp.prefix,
	    conn->nsp.delim);
#endif
	return analyze_response(conn, ibuf.data);
}


/*
 * Process the data that server sent due to IMAP STATUS client request.
 */
int
status_response(conn_t * conn, unsigned int tag, char *mbox)
{
	int r;
	unsigned int exist, recent, unseen;
	char *c;

	exist = recent = unseen = 0;

	reset_buffer(&ibuf);

	do {
		check_buffer(&ibuf, strlen(ibuf.data) + RESPONSE_BUF);
		receive_response(conn, ibuf.data + strlen(ibuf.data));
		bye_response(ibuf.data);
	} while (!strcasestr(ibuf.data, ultostr(tag, 16)));

	r = analyze_response(conn, ibuf.data);

	if (r == RESPONSE_NO)
		return -2;

	if ((c = strcasestr(ibuf.data, "MESSAGES"))) {
		c += 9;
		exist = strtoul(c, NULL, 10);
	}
	if ((c = strcasestr(ibuf.data, "RECENT"))) {
		c += 7;
		recent = strtoul(c, NULL, 10);
	}
	if ((c = strcasestr(ibuf.data, "UNSEEN"))) {
		c += 7;
		unseen = strtoul(c, NULL, 10);
	}
	if (exist == 0) {
		info("No messages in mailbox \"%s\".\n", mbox);
		return -2;
	}
	info("%d message%s, %d recent, %d unseen, in mailbox \"%s\".\n", exist,
	    plural(exist), recent, unseen, mbox);

	return r;
}


/*
 * Process the data that server sent due to IMAP SELECT client request.
 */
int
select_response(conn_t * conn, unsigned int tag)
{
	reset_buffer(&ibuf);

	do {
		check_buffer(&ibuf, strlen(ibuf.data) + RESPONSE_BUF);
		receive_response(conn, ibuf.data + strlen(ibuf.data));
		bye_response(ibuf.data);
	} while (!strcasestr(ibuf.data, ultostr(tag, 16)));

	if (strcasestr(ibuf.data, "[READ-ONLY]"))
		return RESPONSE_READONLY;

	return analyze_response(conn, ibuf.data);
}


/*
 * Process the data that server sent due to IMAP SEARCH client request.
 */
int
search_response(conn_t * conn, unsigned int tag, char **mesgs)
{
	char *c, *m;
	unsigned int blen;

	reset_buffer(&ibuf);

	do {
		check_buffer(&ibuf, strlen(ibuf.data) + RESPONSE_BUF);
		receive_response(conn, ibuf.data + strlen(ibuf.data));
		bye_response(ibuf.data);
	} while (!strcasestr(ibuf.data, ultostr(tag, 16)));

	if ((c = strcasestr(ibuf.data, "* SEARCH "))) {
		blen = strlen(ibuf.data);

		m = *mesgs = (char *)xmalloc(blen + 1);

		c += 9;

		while (*c != '\0' && (isdigit(*c) || *c == ' '))
			*(m++) = *(c++);

		*m = 0;
	}
	return analyze_response(conn, ibuf.data);
}


/*
 * Process the data that server sent due to IMAP FETCH client request.
 */
int
fetch_response(conn_t * conn, unsigned int tag, int reset, char *fetch)
{
	unsigned int i;
	static unsigned int s;
	char *b;

	if (reset) {
		s = 0;
		return 0;
	}
	i = 0;

	reset_buffer(&ibuf);

	do {
		check_buffer(&ibuf, strlen(ibuf.data) + RESPONSE_BUF);
		receive_response(conn, ibuf.data + strlen(ibuf.data));
		bye_response(ibuf.data);
	} while (strlen(ibuf.data) < RESPONSE_BUF &&
	    !strcasestr(ibuf.data, ultostr(tag, 16)));

	b = ibuf.data;

	if (s == 0) {
		if ((b = strstr(b, "}\r\n"))) {
			while (b - ibuf.data > 0)
				if (*--b == '{')
					break;
			s = atoi(++b) - 2;
			b = strchr(b, '}');
			b += 3;
		} else {
			return RESPONSE_NULLBODY;	/* Null body. */
		}
	}
	while (*b != '\0' && s-- != 0)
		fetch[i++] = *(b++);

	fetch[i] = '\0';

	return analyze_response(conn, ibuf.data);
}


/*
 * Process the data that server sent due to IMAP FETCH FAST client request.
 */
int
fetchfast_response(conn_t * conn, char **flags, char **date,
    unsigned int *size, unsigned int tag)
{
	char *c, *m;
	unsigned int blen;

	*size = 0;

	reset_buffer(&ibuf);

	do {
		check_buffer(&ibuf, strlen(ibuf.data) + RESPONSE_BUF);
		receive_response(conn, ibuf.data + strlen(ibuf.data));
		bye_response(ibuf.data);
	} while (!strcasestr(ibuf.data, ultostr(tag, 16)));

	if ((c = strcasestr(ibuf.data, "FETCH (FLAGS ("))) {
		blen = strlen(ibuf.data);

		m = *flags = (char *)xmalloc(blen + 1);

		c += 14;

		while (*c != '\0' && *c != ')') {
			/* The \Recent flag can not be altered by the client. */
			if (*c == '\\') {
				if (strcasestr(c - 1, "(\\Recent)") ||
				    strcasestr(c - 1, " \\Recent)"))
					break;
				else if (strcasestr(c, "\\Recent "))
					c += 8;
			}
			*(m++) = *(c++);
		}

		*m = 0;
	}
	if ((c = strcasestr(ibuf.data, " INTERNALDATE \""))) {
		blen = strlen(ibuf.data);

		m = *date = (char *)xmalloc(blen + 1);

		c += 15;

		while (*c != '\0' && *c != '"')
			*(m++) = *(c++);

		*m = 0;
	}
	if ((c = strcasestr(ibuf.data, " RFC822.SIZE "))) {
		c += 13;
		*size = strtoul(c, NULL, 10);
	}
	return analyze_response(conn, ibuf.data);
}


/*
 * Process the data that server sent due to IMAP APPEND client request.
 */
int
append_response(conn_t * conn, unsigned int tag)
{
	int r;

	r = RESPONSE_OK;

	reset_buffer(&ibuf);

	do {
		check_buffer(&ibuf, strlen(ibuf.data) + RESPONSE_BUF);
		receive_response(conn, ibuf.data + strlen(ibuf.data));
		bye_response(ibuf.data);
	} while (!strcasestr(ibuf.data, ultostr(tag, 16)));

	if ((r = analyze_response(conn, ibuf.data)) == RESPONSE_NO &&
	    strcasestr(ibuf.data, "[TRYCREATE]"))
		return RESPONSE_TRYCREATE;

	return r;
}


/*
 * Process the data that server sent due to IMAP COPY client request.
 */
int
copy_response(conn_t * conn, unsigned int tag)
{
	int r;

	r = RESPONSE_OK;

	reset_buffer(&ibuf);

	do {
		check_buffer(&ibuf, strlen(ibuf.data) + RESPONSE_BUF);
		receive_response(conn, ibuf.data + strlen(ibuf.data));
		bye_response(ibuf.data);
	} while (!strcasestr(ibuf.data, ultostr(tag, 16)));

	if ((r = analyze_response(conn, ibuf.data)) == RESPONSE_NO &&
	    strcasestr(ibuf.data, "[TRYCREATE]"))
		return RESPONSE_TRYCREATE;

	return r;
}


/*
 * Check if response of server to client's request was succesfully
 * delivered or there was some kind of error.
 */
int
analyze_response(conn_t * conn, char *buf)
{
	int r;
	regex_t creg;
	regmatch_t match[3];
	const char *reg;
	char result[RESULT_BUF];

	r = RESPONSE_OK;
	reg = "[[:xdigit:]]{4,4} ((OK|NO|BAD)[[:print:]]*)\r\n";
	result[0] = '\0';

	regcomp(&creg, reg, REG_EXTENDED | REG_ICASE);

	if (!regexec(&creg, buf, 3, match, 0)) {
		strncat(result, buf + match[1].rm_so,
		    min(match[1].rm_eo - match[1].rm_so, RESULT_BUF - 1));

		if (!strncasecmp(buf + match[2].rm_so, "NO", 2))
			r = RESPONSE_NO;
		else if (!strncasecmp(buf + match[2].rm_so, "BAD", 3))
			r = RESPONSE_BAD;

		verbose("%s: %s", (conn == &connpri ? "S" : "s"),
		    buf + match[0].rm_so);
	} else
		r = RESPONSE_NONE;

	regfree(&creg);

	return r;
}
